/* Alain Brobecker, november 2004           */
/*  _ ___          _____          _ ___     */
/* | '   |        |__  -|        | ' __|    */
/* |_|_|_|ultiple |_____|earch & |__|eplace */

#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define FALSE           0
#define TRUE            1

FILE *filep;
unsigned char *Table,*In,*Out; /* data */
int TableLength,InLength,OutLength;
int TablePos;

/**** SearchNext routine, brute force version ******************************/
/* To speed up this routine, copy beforehand the string to search at the   */
/* end of the text (you must allocate a bit more memory) and remove the    */
/* TextPos+StringLength<=TextLength condition in main loop. The routine    */
/* will always find a string, once this is done, compare Pos+Length and    */
/* return -1 if it is the one we added!                                    */
int SearchNext(unsigned char *String,int StringLength,
               unsigned char *Text,int TextLength,int TextPos) {
  int NbMatch=0;
  while((NbMatch<StringLength) && (TextPos+StringLength<=TextLength)) {
    if(Text[TextPos+NbMatch]==String[NbMatch]) {
      NbMatch++;
    } else {
      NbMatch=0;
      TextPos++;
    }
  }
  if(NbMatch==StringLength) {
    return(TextPos);
  } else {
    return(-1);
  }
}

/**** Return value of an hexadecimal char ****/
unsigned char GetHexa(unsigned char c) {
  if((c>='0') && (c<='9')) {
    c=c-'0';
  } else if((c>='a') && (c<='f')) {
    c=c-'a'+10;
  } else if((c>='A') && (c<='F')) {
    c=c-'A'+10;
  } else {
     printf("Wrong hexadecimal number!\n"); exit(1);
  }
  return (c);
}

/**** GetString routine, copy string and return string length ****/
int GetString(unsigned char *String) {
  int end=FALSE;
  int length=0;
  unsigned char c;
  while(end==FALSE) {
    TablePos++;
    if(TablePos==TableLength) { printf("Wrong string!\n"); exit(1); }
    c=Table[TablePos];
    if(c=='#') {
      end=TRUE;
    } else {
      if(c=='^') {
        TablePos++;
        if(TablePos==TableLength) { printf("Wrong string!\n"); exit(1); }
        c=Table[TablePos];
        if((c!='#') && (c!='^')) {
          c=GetHexa(Table[TablePos]);
          TablePos++;
          if(TablePos==TableLength) { printf("Wrong string!\n"); exit(1); }
          c=c*16+GetHexa(Table[TablePos]);
        }
      }
      String[length]=c;
      length++;
    }
  }
  String[length]='\0';
  return(length);
}

/**** Main *****************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  unsigned char *SS,*RS; /* SS=SearchString, RS=ReplaceString */
  int SSLength,RSLength,NbSS; /* Lengths and number of SS */
  int InPos,NextPos,OutPos;
  int loop,c;

  if(argc!=4) { printf("usage: MSR TableFile InFile OutFile\n"); exit(1); }

  /**** load TableFile ****/
  filep=fopen(argv[1],"rb");
  if(filep==0) { printf("error: %s not found\n",argv[1]); exit(1); }
  fseek(filep,0,2); /* 2 for EOF */
  TableLength=(int) ftell(filep);
  fseek(filep,0,0); /* 0 for SOF */
  Table=(unsigned char *) malloc((unsigned int) TableLength);
  fread(Table,1,TableLength,filep);
  fclose(filep);

  /**** Search & Replace string are smaller than Table ****/
  SS=(unsigned char *) malloc((unsigned int) TableLength);
  RS=(unsigned char *) malloc((unsigned int) TableLength);

  /**** load InFile ****/
  filep=fopen(argv[2],"rb");
  if(filep==0) { printf("error: %s not found\n",argv[1]); exit(1); }
  fseek(filep,0,2); /* 2 for EOF */
  InLength=(int) ftell(filep);
  fseek(filep,0,0); /* 0 for SOF */
  In=(unsigned char *) malloc((unsigned int) InLength);
  fread(In,1,InLength,filep);
  fclose(filep);

  /**** S&R routine starts here! ****/
  TablePos=SearchNext((unsigned char *) "#",1,Table,TableLength,0);
  while(TablePos!=-1) {
    SSLength=GetString(SS); /* Get SS */
    RSLength=GetString(RS); /* Get RS */
    TablePos++;
    loop=FALSE;
    /* If replace string is finished by #! then we must loop */
    if((TablePos!=TableLength) && (Table[TablePos]=='!')) { loop=TRUE; }
    do {
      /* Count the numberb of SS in InFile */
      InPos=0;
      NbSS=0;
      do {
        InPos=SearchNext(SS,SSLength,In,InLength,InPos);
        if(InPos!=-1) {
          NbSS++;
          InPos=InPos+SSLength;
        }
      } while(InPos!=-1);
      /* If there are occurences of SS, allocate memory and replace them */
      if(NbSS!=0) {
        OutLength=InLength+NbSS*(RSLength-SSLength);
        Out=(unsigned char *) malloc((unsigned int) OutLength);
        InPos=0;
        OutPos=0;
        do {
          NextPos=SearchNext(SS,SSLength,In,InLength,InPos);
          if(NextPos==-1) { /* No more SS, copy rest of file */
            for(c=InPos;c<InLength;c++) {
              Out[OutPos]=In[c];
              OutPos++;
            }
          } else { /* One SS found, copy before string then replace it */
            for(c=InPos;c<NextPos;c++) { Out[OutPos++]=In[c]; }
            for(c=0;c<RSLength;c++) { Out[OutPos++]=RS[c]; }
            InPos=NextPos+SSLength; /* point after this occurence of SS */
          }
        } while(NextPos!=-1);
        free(In); /* remove In text */
        In=Out;
        InLength=OutPos;
      } else {
        loop=FALSE; /* No more SS, so make sure we don't loop anymore */
      }
    } while(loop==TRUE);
    /* Search next command */
    TablePos=SearchNext((unsigned char *) "#",1,Table,TableLength,TablePos);
  }

  /**** save resulting file ****/
  filep=fopen(argv[3],"wb");
  fwrite(In,1,InLength,filep);
  fclose(filep);
}
